@@ -27,7 +27,6 @@ class Mega(object):
             options = {}
         self.options = options
 
-
     def login(self, email, password):
         self.login_user(email, password)
         return self
@@ -262,7 +261,12 @@ class Mega(object):
         """
         Get all files in a given target, e.g. 4=trash
         """
-        node_id = self.get_node_by_type(target)
+        if type(target) == int:
+            # convert special nodes (e.g. trash)
+            node_id = self.get_node_by_type(target)
+        else:
+            node_id = [target]
+
         files = self.api_request({'a': 'f', 'c': 1})
         files_dict = {}
         shared_keys = {}
@@ -275,11 +279,16 @@ class Mega(object):
 
     def get_id_from_public_handle(self, public_handle):
         #get node data
-        #TODO fix this function
         node_data = self.api_request({'a': 'f', 'f': 1, 'p': public_handle})
+        node_id = self.get_id_from_obj(node_data)
+        return node_id
+
+    def get_id_from_obj(self, node_data):
+        """
+        Get node id from a file object
+        """
         node_id = None
 
-        #determine node id
         for i in node_data['f']:
             if i['h'] is not u'':
                 node_id = i['h']
@@ -305,16 +314,16 @@ class Mega(object):
             raise ValueError("Only one unit prefix can be specified")
         unit_coef = 1
         if kilo:
-          unit_coef = 1024
+            unit_coef = 1024
         if mega:
-          unit_coef = 1048576
+            unit_coef = 1048576
         if giga:
-          unit_coef = 1073741824
+            unit_coef = 1073741824
         json_resp = self.api_request({'a': 'uq', 'xfer': 1, 'strg': 1})
         return {
             'used': json_resp['cstrg'] / unit_coef,
             'total': json_resp['mstrg'] / unit_coef,
-            }
+        }
 
     def get_balance(self):
         """
@@ -327,51 +336,37 @@ class Mega(object):
     ##########################################################################
     # DELETE
     def delete(self, public_handle):
-        #straight delete by id
+        """
+        Delete a file by its public handle
+        """
         return self.move(public_handle, 4)
 
     def delete_url(self, url):
-        #delete a file via it's url
+        """
+        Delete a file by its url
+        """
         path = self.parse_url(url).split('!')
         public_handle = path[0]
         file_id = self.get_id_from_public_handle(public_handle)
         return self.move(file_id, 4)
 
     def destroy(self, file_id):
-        #delete forever by private id
+        """
+        Destroy a file by its private id
+        """
         return self.api_request({'a': 'd',
                                  'n': file_id,
                                  'i': self.request_id})
 
     def destroy_url(self, url):
-        #delete a file via it's url
+        """
+        Destroy a file by its url
+        """
         path = self.parse_url(url).split('!')
         public_handle = path[0]
         file_id = self.get_id_from_public_handle(public_handle)
         return self.destroy(file_id)
 
-    def move(self, file_id, target):
-        """
-        Move a file to another parent node
-        params:
-        a : command
-        n : node we're moving
-        t : id of target parent node, moving to
-        i : request id
-
-        targets
-        2 : root
-        3 : inbox
-        4 : trash
-        """
-
-        #determine target_node_id
-        target_node_id = str(self.get_node_by_type(target)[0])
-        return self.api_request({'a': 'm',
-                                 'n': file_id,
-                                 't': target_node_id,
-                                 'i': self.request_id})
-
     def empty_trash(self):
         # get list of files in rubbish out
         files = self.get_files_in_node(4)
@@ -387,22 +382,22 @@ class Mega(object):
 
     ##########################################################################
     # DOWNLOAD
-    def download(self, file, dest_path=None):
+    def download(self, file, dest_path=None, dest_filename=None):
         """
         Download a file by it's file object
         """
-        self.download_file(None, None, file=file[1], dest_path=dest_path, is_public=False)
+        self.download_file(None, None, file=file[1], dest_path=dest_path, dest_filename=dest_filename, is_public=False)
 
-    def download_url(self, url, dest_path=None):
+    def download_url(self, url, dest_path=None, dest_filename=None):
         """
         Download a file by it's public url
         """
         path = self.parse_url(url).split('!')
         file_id = path[0]
         file_key = path[1]
-        self.download_file(file_id, file_key, dest_path, is_public=True)
+        self.download_file(file_id, file_key, dest_path, dest_filename, is_public=True)
 
-    def download_file(self, file_handle, file_key, dest_path=None, is_public=False, file=None):
+    def download_file(self, file_handle, file_key, dest_path=None, dest_filename=None, is_public=False, file=None):
         if file is None :
             if is_public:
                 file_key = base64_to_a32(file_key)
@@ -429,7 +424,11 @@ class Mega(object):
         file_size = file_data['s']
         attribs = base64_url_decode(file_data['at'])
         attribs = decrypt_attr(attribs, k)
-        file_name = attribs['n']
+
+        if dest_filename is not None:
+            file_name = dest_filename
+        else:
+            file_name = attribs['n']
 
         input_file = requests.get(file_url, stream=True).raw
 
@@ -440,41 +439,43 @@ class Mega(object):
 
         temp_output_file = tempfile.NamedTemporaryFile(mode='w+b', prefix='megapy_', delete=False)
 
+        k_str = a32_to_str(k)
         counter = Counter.new(
             128, initial_value=((iv[0] << 32) + iv[1]) << 64)
-        aes = AES.new(a32_to_str(k), AES.MODE_CTR, counter=counter)
+        aes = AES.new(k_str, AES.MODE_CTR, counter=counter)
 
-        file_mac = (0, 0, 0, 0)
-        for chunk_start, chunk_size in sorted(get_chunks(file_size).items()):
+        mac_str = '\0' * 16
+        mac_encryptor = AES.new(k_str, AES.MODE_CBC, mac_str)
+        iv_str = a32_to_str([iv[0], iv[1], iv[0], iv[1]])
+
+        for chunk_start, chunk_size in get_chunks(file_size):
             chunk = input_file.read(chunk_size)
             chunk = aes.decrypt(chunk)
             temp_output_file.write(chunk)
 
-            chunk_mac = [iv[0], iv[1], iv[0], iv[1]]
-            for i in range(0, len(chunk), 16):
+            encryptor = AES.new(k_str, AES.MODE_CBC, iv_str)
+            for i in range(0, len(chunk)-16, 16):
                 block = chunk[i:i + 16]
-                if len(block) % 16:
-                    block += '\0' * (16 - (len(block) % 16))
-                block = str_to_a32(block)
-                chunk_mac = [
-                    chunk_mac[0] ^ block[0],
-                    chunk_mac[1] ^ block[1],
-                    chunk_mac[2] ^ block[2],
-                    chunk_mac[3] ^ block[3]]
-                chunk_mac = aes_cbc_encrypt_a32(chunk_mac, k)
+                encryptor.encrypt(block)
 
-            file_mac = [
-                file_mac[0] ^ chunk_mac[0],
-                file_mac[1] ^ chunk_mac[1],
-                file_mac[2] ^ chunk_mac[2],
-                file_mac[3] ^ chunk_mac[3]]
-            file_mac = aes_cbc_encrypt_a32(file_mac, k)
+            #fix for files under 16 bytes failing
+            if file_size > 16:
+                i += 16
+            else:
+                i = 0
+
+            block = chunk[i:i + 16]
+            if len(block) % 16:
+                block += '\0' * (16 - (len(block) % 16))
+            mac_str = mac_encryptor.encrypt(encryptor.encrypt(block))
 
             if self.options.get('verbose') is True:
                 # temp file size
                 file_info = os.stat(temp_output_file.name)
                 print('{0} of {1} downloaded'.format(file_info.st_size, file_size))
 
+        file_mac = str_to_a32(mac_str)
+
         temp_output_file.close()
 
         # check mac integrity
@@ -485,7 +486,7 @@ class Mega(object):
 
     ##########################################################################
     # UPLOAD
-    def upload(self, filename, dest=None):
+    def upload(self, filename, dest=None, dest_filename=None):
         #determine storage node
         if dest is None:
             #if none set, upload to cloud drive node
@@ -500,32 +501,36 @@ class Mega(object):
 
         #generate random aes key (128) for file
         ul_key = [random.randint(0, 0xFFFFFFFF) for _ in range(6)]
+        k_str = a32_to_str(ul_key[:4])
         count = Counter.new(128, initial_value=((ul_key[4] << 32) + ul_key[5]) << 64)
-        aes = AES.new(a32_to_str(ul_key[:4]), AES.MODE_CTR, counter=count)
+        aes = AES.new(k_str, AES.MODE_CTR, counter=count)
 
-        file_mac = [0, 0, 0, 0]
         upload_progress = 0
-        for chunk_start, chunk_size in sorted(get_chunks(file_size).items()):
+        completion_file_handle = None
+
+        mac_str = '\0' * 16
+        mac_encryptor = AES.new(k_str, AES.MODE_CBC, mac_str)
+        iv_str = a32_to_str([ul_key[4], ul_key[5], ul_key[4], ul_key[5]])
+
+        for chunk_start, chunk_size in get_chunks(file_size):
             chunk = input_file.read(chunk_size)
             upload_progress += len(chunk)
 
-            #determine chunks mac
-            chunk_mac = [ul_key[4], ul_key[5], ul_key[4], ul_key[5]]
-            for i in range(0, len(chunk), 16):
+            encryptor = AES.new(k_str, AES.MODE_CBC, iv_str)
+            for i in range(0, len(chunk)-16, 16):
                 block = chunk[i:i + 16]
-                if len(block) % 16:
-                    block += '\0' * (16 - len(block) % 16)
-                block = str_to_a32(block)
-                chunk_mac = [chunk_mac[0] ^ block[0], chunk_mac[1] ^ block[1],
-                             chunk_mac[2] ^ block[2],
-                             chunk_mac[3] ^ block[3]]
-                chunk_mac = aes_cbc_encrypt_a32(chunk_mac, ul_key[:4])
+                encryptor.encrypt(block)
 
-            #our files mac
-            file_mac = [file_mac[0] ^ chunk_mac[0], file_mac[1] ^ chunk_mac[1],
-                        file_mac[2] ^ chunk_mac[2],
-                        file_mac[3] ^ chunk_mac[3]]
-            file_mac = aes_cbc_encrypt_a32(file_mac, ul_key[:4])
+            #fix for files under 16 bytes failing
+            if file_size > 16:
+                i += 16
+            else:
+                i = 0
+
+            block = chunk[i:i + 16]
+            if len(block) % 16:
+                block += '\0' * (16 - len(block) % 16)
+            mac_str = mac_encryptor.encrypt(encryptor.encrypt(block))
 
             #encrypt file and upload
             chunk = aes.encrypt(chunk)
@@ -537,10 +542,16 @@ class Mega(object):
                 # upload progress
                 print('{0} of {1} uploaded'.format(upload_progress, file_size))
 
+        file_mac = str_to_a32(mac_str)
+
         #determine meta mac
         meta_mac = (file_mac[0] ^ file_mac[1], file_mac[2] ^ file_mac[3])
 
-        attribs = {'n': os.path.basename(filename)}
+        if dest_filename is not None:
+            attribs = {'n': dest_filename}
+        else:
+            attribs = {'n': os.path.basename(filename)}
+
         encrypt_attribs = base64_url_encode(encrypt_attr(attribs, ul_key[:4]))
         key = [ul_key[0] ^ ul_key[4], ul_key[1] ^ ul_key[5],
                ul_key[2] ^ meta_mac[0], ul_key[3] ^ meta_mac[1],
@@ -556,6 +567,8 @@ class Mega(object):
         input_file.close()
         return data
 
+    ##########################################################################
+    # OTHER OPERATIONS
     def create_folder(self, name, dest=None):
         #determine storage node
         if dest is None:
@@ -584,3 +597,179 @@ class Mega(object):
                                  'i': self.request_id})
         #return API msg
         return data
+
+    def rename(self, file, new_name):
+        file = file[1]
+        #create new attribs
+        attribs = {'n': new_name}
+        #encrypt attribs
+        encrypt_attribs = base64_url_encode(encrypt_attr(attribs, file['k']))
+        encrypted_key = a32_to_base64(encrypt_key(file['key'], self.master_key))
+
+        #update attributes
+        data = self.api_request([{
+            'a': 'a',
+            'attr': encrypt_attribs,
+            'key': encrypted_key,
+            'n': file['h'],
+            'i': self.request_id}])
+
+        #return API msg
+        return data
+
+    def move(self, file_id, target):
+        """
+        Move a file to another parent node
+        params:
+        a : command
+        n : node we're moving
+        t : id of target parent node, moving to
+        i : request id
+
+        targets
+        2 : root
+        3 : inbox
+        4 : trash
+
+        or...
+        target's id
+        or...
+        target's structure returned by find()
+        """
+
+        #determine target_node_id
+        if type(target) == int:
+            target_node_id = str(self.get_node_by_type(target)[0])
+        elif type(target) in (str, unicode):
+            target_node_id = target
+        else:
+            file = target[1]
+            target_node_id = file['h']
+        return self.api_request({'a': 'm',
+                                 'n': file_id,
+                                 't': target_node_id,
+                                 'i': self.request_id})
+
+    def add_contact(self, email):
+        """
+        Add another user to your mega contact list
+        """
+        return self._edit_contact(email, True)
+
+    def remove_contact(self, email):
+        """
+        Remove a user to your mega contact list
+        """
+        return self._edit_contact(email, False)
+
+    def _edit_contact(self, email, add):
+        """
+        Editing contacts
+        """
+        if add is True:
+            l = '1'  # add command
+        elif add is False:
+            l = '0'  # remove command
+        else:
+            raise ValidationError('add parameter must be of type bool')
+
+        if not re.match(r"[^@]+@[^@]+\.[^@]+", email):
+            ValidationError('add_contact requires a valid email address')
+        else:
+            return self.api_request({'a': 'ur',
+                                    'u': email,
+                                    'l': l,
+                                    'i': self.request_id})
+
+    def get_contacts(self):
+        raise NotImplementedError()
+        # TODO implement this
+        # sn param below = maxaction var with function getsc() in mega.co.nz js
+        # seens to be the 'sn' attrib of the previous request response...
+        # mega.co.nz js full source @ http://homepages.shu.ac.uk/~rjodwyer/mega-scripts-all.js
+        # requests goto /sc rather than
+
+        #req = requests.post(
+        #'{0}://g.api.{1}/sc'.format(self.schema, self.domain),
+        #    params={'sn': 'ZMxcQ_DmHnM', 'ssl': '1'},
+        #    data=json.dumps(None),
+        #    timeout=self.timeout)
+        #json_resp = json.loads(req.text)
+        #print json_resp
+    
+    def get_public_url_info(self, url):
+        """
+        Get size and name from a public url, dict returned
+        """
+        file_handle, file_key = self.parse_url(url).split('!')
+        return self.get_public_file_info(file_handle, file_key)
+
+    def import_public_url(self, url, dest_node=None, dest_name=None):
+        """
+        Import the public url into user account
+        """
+        file_handle, file_key = self.parse_url(url).split('!')
+        return self.import_public_file(file_handle, file_key, dest_node=dest_node, dest_name=dest_name)
+
+    def get_public_file_info(self, file_handle, file_key):
+        """
+        Get size and name of a public file
+        """
+        data = self.api_request({
+            'a': 'g',
+            'p': file_handle,
+            'ssm': 1})
+
+        #if numeric error code response
+        if isinstance(data, int):
+            raise RequestError(data)
+
+        if 'at' not in data or 's' not in data:
+            raise ValueError("Unexpected result", data)
+
+        key = base64_to_a32(file_key)
+        k = (key[0] ^ key[4], key[1] ^ key[5], key[2] ^ key[6], key[3] ^ key[7])
+
+        size = data['s']
+        unencrypted_attrs = decrypt_attr(base64_url_decode(data['at']), k)
+        if not(unencrypted_attrs):
+            return None
+
+        result = {
+            'size': size,
+            'name': unencrypted_attrs['n']}
+
+        return result
+
+    def import_public_file(self, file_handle, file_key, dest_node=None, dest_name=None):
+        """
+        Import the public file into user account
+        """
+
+        # Providing dest_node spare an API call to retrieve it.
+        if dest_node is None:
+            # Get '/Cloud Drive' folder no dest node specified
+            dest_node = self.get_node_by_type(2)[1]
+
+        # Providing dest_name spares an API call to retrieve it.
+        if dest_name is None:
+            pl_info = self.get_public_file_info(file_handle, file_key)
+            dest_name = pl_info['name']
+
+        key = base64_to_a32(file_key)
+        k = (key[0] ^ key[4], key[1] ^ key[5], key[2] ^ key[6], key[3] ^ key[7])
+
+        encrypted_key = a32_to_base64(encrypt_key(key, self.master_key))
+        encrypted_name = base64_url_encode(encrypt_attr({'n': dest_name}, k))
+
+        data = self.api_request({
+            'a': 'p',
+            't': dest_node['h'],
+            'n': [{
+                'ph': file_handle,
+                't': 0,
+                'a': encrypted_name,
+                'k': encrypted_key}]})
+
+        #return API msg
+        return data
